'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

// Given a pipeline, collapses it to a filter, a window and an ordering
let compile = (() => {
  var _ref2 = _asyncToGenerator(function* (scope, pipeline, executor) {
    const operations = pipeline.operations;
    if (operations.length == 0) {
      throw new Error("Empty pipelines can't be collapsed");
    }
    if (operations[0].op != 'source') {
      throw new Error("Can't collapse pipelines that does not start with an *");
    }

    const sourceId = operations[0].id;

    const fetchSpec = new FetchSpec(scope, sourceId);
    const ops = operations.slice(1);
    while (ops.length > 0) {
      const operation = ops.pop();
      (0, _debug2.default)('compiling', operation);
      switch (operation.op) {
        case 'filter':
          fetchSpec.applyFilter(operation.filter);
          break;
        case 'mapJoin':
          const filter = yield generalizeMapJoin(operation, scope, executor);
          fetchSpec.applyFilter(filter);
        case 'object':
          fetchSpec.project(operation.operations);
          break;
        case 'subscript':
          fetchSpec.applyWindow(operation.start, operation.end);
          break;
        case 'ordering':
          fetchSpec.applyOrdering(operation.terms);
        case 'accessor':
          fetchSpec.project(operation.operations);
          break;
        default:
          throw new Error(`Unknown pipeline operation ${operation.op} when compiling fetch-operation`);
      }
    }

    const generalizedFilter = yield (0, _generalizeJoinFilter2.default)(fetchSpec.filter, scope);
    const wasJoin = generalizedFilter !== fetchSpec.filter;

    // If this was a join, replace fetch with a join-fetch
    if (wasJoin) {
      const joinFetch = new FetchSpec(scope, sourceId);
      joinFetch.filter = generalizedFilter;
      return joinFetch;
    }

    // TODO: Now the filter needs to be expanded in case it requires anything
    // to be fetched in order to be executed.
    (0, _debug2.default)('fetchSpec => ', fetchSpec);
    return fetchSpec;
  });

  return function compile(_x5, _x6, _x7) {
    return _ref2.apply(this, arguments);
  };
})();

// Given a map join, generates a filter that will fetch all possible source values for that
// join.


let generalizeMapJoin = (() => {
  var _ref3 = _asyncToGenerator(function* (operation, scope, executor) {
    if (!scope) {
      (0, _debug2.default)("mapJoin over null short circuited");
      fetchSpec.applyFilter(new _operations.Literal({
        type: 'boolean',
        value: false
      }));
    }
    let ids;
    if (scope.sourceId) {
      const source = yield scope.dataForSource(scope.sourceId);
      (0, _debug2.default)('expanding mapJoin', scope, scope.sourceId);
      const idPromises = source.documents.map((() => {
        var _ref4 = _asyncToGenerator(function* (document) {
          const scope = new _Scope2.default({
            value: document
          });
          const resolved = (0, _scopeTools.asPlainValue)((yield executor.exec(operation.pipeline, scope))).filter(Boolean);
          (0, _debug2.default)('resolved', resolved);
          if (resolved) {
            return resolved;
          }
        });

        return function (_x11) {
          return _ref4.apply(this, arguments);
        };
      })());
      const idMap = {};
      ids = (yield Promise.all(idPromises)).forEach(function (idSet) {
        return idSet.forEach(function (id) {
          idMap[id] = true;
        });
      });
      ids = Object.keys(idMap);
    } else {
      // TODO
      throw new Error(`mapJoin on literals not supported yet`);
    }
    (0, _debug2.default)('expanded mapJoin', ids);
    return new _operations.BinaryOperator('in', new _operations.Accessor([new _operations.Attribute('_id')]), new _operations.Literal({
      value: ids
    }));
  });

  return function generalizeMapJoin(_x8, _x9, _x10) {
    return _ref3.apply(this, arguments);
  };
})();

var _operations = require('../plan/operations');

var _Scope = require('./Scope');

var _Scope2 = _interopRequireDefault(_Scope);

var _debug = require('../debug');

var _debug2 = _interopRequireDefault(_debug);

var _generalizeJoinFilter = require('./generalizeJoinFilter');

var _generalizeJoinFilter2 = _interopRequireDefault(_generalizeJoinFilter);

var _scopeTools = require('./scopeTools');

var _path = require('path');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

// Given a pipeline, fetch will make sure the source value for that pipeline
// gets cached in the parentScope. It does not actually return the retrieved
// value, but the same value with the fetch result cached into the sourceId
// of the provided pipeline.
// The fetcher is expected to return an object that looks something like this
// {
//   results: [{_id: 'doc4', _type: 'document'}, {_id: 'doc5', _type: 'document'}, ref: {_ref: 'doc4', weak: true}],
//   start: 3,
//   refs: {'doc5': [{id: 'doc3', weak: true}]}
// }
// `results` is required, `start` is used to signal that this set is allready offset by the fetcher so that the
// execution engine will not double-offset it, `refs` is a flattened list of all refs in every included document.
// This is optional and will be generated on demand if missing.

exports.default = (() => {
  var _ref = _asyncToGenerator(function* (scope, pipeline, fetcher, executor) {
    (0, _debug2.default)('fetch()', scope, pipeline);
    const sourceId = extractSourceId(pipeline);
    // Is it a fetchable source, and did we not fetch it allready?
    if (!sourceId || scope.sourceIsCached(sourceId)) {
      (0, _debug2.default)('fetch() short circuit', sourceId, scope.sourceIsCached(sourceId));
      return scope;
    }
    // We need to actually go get it
    const responder = scope.claimSource(sourceId);
    const fetchSpec = yield compile(scope, pipeline, executor);
    const response = yield fetcher(fetchSpec);
    // Cache the result on the scope so it will bind to the source
    // when it comes up later
    (0, _debug2.default)('fetch() =>', response);
    scope.cacheRefs(response.refs);
    responder.resolve({
      documents: response.results,
      start: response.start || 0
    });
    return scope;
  });

  function fetch(_x, _x2, _x3, _x4) {
    return _ref.apply(this, arguments);
  }

  return fetch;
})();

// Keeps track of the central parameters of a fetch operation.
// Since pipelines are computed in reverse order, filters,
// orderings and windows are expected to be applied in reverse order
// last to first.


class FetchSpec {
  constructor(parentScope, sourceId) {
    this.parentScope = parentScope;
    this.sourceId = sourceId;
    this.clear();
  }

  applyFilter(constraint) {
    if (!this.filter) {
      this.filter = constraint;
      return;
    }
    this.filter = new _operations.BinaryOperator('and', this.filter, constraint);
  }

  project(objectOperations) {
    // TODO: Remap the values of the filter and orderings collected so far
    // by substituting expressions according to the projection
  }

  applyWindow(start, end) {
    (0, _debug2.default)("applyWindow()", start, end);
    if (!Number.isFinite(start) || !Number.isFinite(end)) {
      throw new Error("start and end must be finite when applying window");
    }
    if (this.start === null) {
      this.start = start;
      this.end = end;
      return;
    }
    this.start += start;
    this.end = Math.min(this.end + start, end);
  }

  applyOrdering(terms) {
    this.ordering = terms.concat(this.ordering);
  }

  clear() {
    this.filter = null;
    this.start = null;
    this.end = null;
    this.ordering = [];
  }
}

function extractSourceId(pipeline) {
  const operations = pipeline.operations;
  if (operations.length == 0) {
    (0, _debug2.default)('extractSourceId: empty pipeline');
    return null;
  }
  if (operations[0].op != 'source') {
    (0, _debug2.default)('extractSourceId: op 0 is not source', operations[0]);
    return null;
  }
  return operations[0].id;
}
//# sourceMappingURL=fetch.js.map